------------------------------------------------------
--               Merseanne Twister                  --
------------------------------------------------------
-- A VERY GOOD random number generator!! Check
-- http://http://www.math.sci.hiroshima-u.ac.jp/m-mat/MT/MT2002/emt19937ar.html
-- or any other web page for more information:
--
-- This is a translation of MASM assembler code
-- to euphoria with some modifications to make it work.
-- It runs on DOS, LINUX and WINDOWS
--
-- The original version was obtained from: http://www.agner.org/random/
-- It's a highly optimized version of this RNG,
-- It includes optimizations for MMX and SSE2, but I
-- didn't include them since my computer doesn't support them.
-- I also didn't include tempering...actually I don't know what that is...
-- because the original code indicates that this decreases speed by 25%
--
-- Performance tests show that the speed is comparable to rand() (just a bit slower)
-- when running from the EU interpreter and about TWICE as fast
-- when compiling through C.
--
-- I haven't run any rigorous test to ensure this version works fine,
-- but the small Monte Carlo simulations that I have run with this RNG
-- indicate that everything is OK.
--
-- * InitMT(seed)  --Run once at the beginning of the program
--               --It calls TRandomInit(seed) to generate a new seed
-- * TRandomInit(seed) --Starts the RNG with a new seed. You don't need to
--                   --call it if you don't want to change the seed and
--                   --InitMT has been called once.
-- * TIRandom(min,max)  --returns an integer in the interval [min,max]
-- * TRandom()          --returns a 32bit floating point number in the
--                    --range [0,1)
--                    --This version is some how slow because it uses
--                    --float64_to_atom routine, I couldn't find a better way to
--                    --do it. Use J_Random() preferably.
-- * J_Random()         --This is my faster version of TRandom()
--                    --Use this one instead.
-- * Free_MT()          --Frees some allocated memory. You should call it at the end
--                    --of your program though I think memory will be reclaimed by
--                    --the OS anyway
--
-- Juan Reina  Oct 2004    lousyshredder@yahoo.com
--
--

--without type_check
--without trace

include machine.e
include asm.e
--include asmdebug.e

sequence Float32,Float64
atom INIT1,TRandomInit_f,TRandom_f,TBRandom_f,TIRandom_f,J_TIRandom_f
atom MT,MTI,seed,FLREADY,FLOAT32,max,min,returnvalue

--
-- A simple pause
--
procedure pause()
  while (get_key()=-1) do end while
end procedure

--
-- Gets machine code for Random Functions
--
procedure Init_F()

 INIT1 = get_asm(
 --"INIT1: \n"&  -- make random seeds from eax and put them into MT buffer
      "xor ecx,ecx\n"&
   "R10:    \n"&
      "mov [MT+ecx*4],eax\n"&
      "mov edx,eax\n"&
      "shr eax,30\n"&
      "xor eax,edx\n"&
      "db #69 db #C0 db #65 db #89 db #07 db #6C\n"&   --"imul eax,#6C078965\n"&
      "inc ecx\n"&
      "add eax,ecx\n"&
      "cmp ecx,624\n"& --***
      "jb R10\n"&
      "shl ecx,2\n"&
      "mov [MTI],ecx\n"&
      "ret")
  resolve_param("MT",MT)
  resolve_param("MTI",MTI)

 TRandomInit_f = get_asm((--debug_asm(
      "pusha\n"&
      "mov eax,[seed]\n"&        -- seed
      "mov ebx,INIT1\n"&
      "call near ebx\n"&     -- initialize MT buffer with seeds
      "popa\n"&
      "ret\n"))
  resolve_param("seed",seed)
  resolve_param("INIT1",INIT1)

  TBRandom_f = get_asm((--debug_asm(
-- this version is for old processors without XMM support:
   "R20:\n"&
	"mov ecx,[MTI]\n"&
	"cmp ecx,2496\n"&   --N*4 = 624*4
	"jnb near R50\n"&   --buffer is empty, fill it
    "R40:\n"&
	"mov eax,[MT+ecx]\n"&
	"add ecx,4\n"&
	"mov [MTI],ecx\n"&
	"db #C3\n"&      --ret: cheating, bug in asm.e?
	-- fill buffer with random numbers
    "R50:\n"&
	"push ebx\n"&
	"mov ecx,0\n"&      --offset MT
	"mov ebx,1588\n"&  --offset MT + M*4
	-- kk loop
    "R60:\n"&
	"mov eax,[MT+ecx]\n"&
	"mov edx,[MT+ecx+4]\n"&
	"and eax,#80000000\n"&     --UPPER_MASK = -1 SHL R   --R=19
	"and edx,#7FFFFFFF\n"&        --LOWER_MASK = (1 SHL R) - 1
	"or eax,edx\n"&
	"shr eax,1\n"&
	"sbb edx,edx\n"&
	"and edx,#9908B0DF\n"&
	"xor eax,edx\n"&
	"xor eax,[MT+ebx]\n"&
	"mov [MT+ecx],eax\n"&
	"add ebx,4\n"&
	"cmp ebx,2496\n"&  --offset MT + N*4 --M=1404
	"jb near R70\n"&
	"mov eax,[MT]\n"&
	"mov [MT+ebx],eax\n"&  -- copy begin of table to after end to simplify kk+1 wraparound
	"mov ebx,0\n"&   --offset MT
    "R70:\n"&
       "add ecx,4\n"&
       "cmp ecx,2496\n"& --offset MT + N*4
       "jb near R60\n"&  -- loop end
       "xor ecx,ecx\n"&
       "mov [MTI],ecx\n"&
       "pop ebx\n"&
       "jmp near R40\n")) --TBRandom_f
   resolve_param("MT",MT)
   resolve_param("MTI",MTI)


TRandom_f = get_asm((--debug_asm(
--     "pusha pushf\n"&
     "test BYTE PTR [FLREADY],1\n"&      -- do we have a number ready to use?
     "jnz near TR10\n"&
  "TR80:\n"&                    -- first-time call: no premade number is ready
     "call near TR20\n"&        -- make a random number
     "or BYTE PTR [FLREADY],1\n"&        -- remember that we have a number
  "TR10:\n"&                  -- a number is already ready
     --"fld QWORD PTR [FLOAT32]\n"&       -- premade random number in interval [1,2)
     --"mov ebx,#232332\n"&
     --"fsub ebx\n"&          -- adjust interval
  "TR20:\n"&                  -- make a new number ready for next time
     "mov ebx,TBRandom\n"&
     "call near ebx\n"&     -- random bits
     "mov edx,eax\n"&            -- make a random number ready for next time
     "shr eax,12\n"&
     "or eax,#3FF00000\n"&
     "shl edx,20\n"&
     "mov dword ptr [FLOAT32+4],eax\n"&
     "mov dword ptr [FLOAT32],edx\n"&
--   "ftsp qword ptr [ReturnFloat]\n"&
--   "popf popa\n"&
     "ret\n"))
  resolve_param("TBRandom",TBRandom_f)
  resolve_param("FLREADY",FLREADY)
  resolve_param("FLOAT32",FLOAT32)

TIRandom_f = get_asm(
       "pusha\n"&
       "mov ebx,TBRandom\n"&
       "call near ebx\n"&   --make random number
       "mov edx,[max]\n"&         --max
       "mov ecx,[min]\n"&         -- min
       "sub edx,ecx\n"&
       "js near RERROR\n"&      -- max < min
       "add edx,1\n"&           -- max - min + 1
       "mul edx\n"&             -- multiply random number by interval and truncate
       "lea eax,[edx+ecx]\n"&   -- add min
       "mov dword ptr [return],eax\n"&
       "popa\n"&
       "db #C3\n"&
"RERROR: \n"&
	"mov eax,#80000000\n"&  -- error exit
	"mov [return],eax\n"&
	"popa\n"&
	"db #C3\n")
  resolve_param("TBRandom",TBRandom_f)
  resolve_param("max",max)
  resolve_param("min",min)
  resolve_param("return",returnvalue)

--This one is a modification of TIRandom for my routine J_Random()
J_TIRandom_f = get_asm(
       "pusha\n"&
       "mov ebx,TBRandom\n"&
       "call near ebx\n"&   --make random number
       "mov edx,1073741823\n"&
       "mul edx\n"&             -- multiply random number by interval and truncate
       "lea eax,[edx]\n"&   -- add min
       "mov dword ptr [return],eax\n"&
       "popa\n"&
       "ret\n")
  resolve_param("TBRandom",TBRandom_f)
  resolve_param("return",returnvalue)


end procedure --Init_F


-------GLOBAL PROCEDURES AND FUNCTIONS-----

--
-- Releases memory
--
global procedure Free_MT()

--asm_cleanup()
 free(MT)
 free(MTI)
 free(seed)
 free(FLREADY)
 free(FLOAT32)
 free(max)
 free(min)

end procedure

--
--Sets random seed
--
global procedure TRandomInit(integer randseed)
    poke4(seed,randseed)
    call(TRandomInit_f)
end procedure


--
-- Allocates memory variables and machine code of functions
--
global procedure InitMT(integer randseed)

MT = allocate((624+4)*4)      poke(MT,repeat(0,(624+4)*4))
MTI = allocate(4)             poke4(MTI,0)
seed = allocate(4)            poke4(seed,12345)

FLREADY = allocate(1)         poke(FLREADY,0)
FLOAT32 = allocate(8)         poke4(FLOAT32,0) poke4(FLOAT32+4,0) --QWORD

Float64 = repeat(0,8)
Float32 = repeat(0,8)

max = allocate(4)             poke4(max,#7FFFF)
min = allocate(4)             poke4(min,1)
returnvalue = allocate(4)     poke4(returnvalue,0)

Init_F()
TRandomInit(randseed)
end procedure

--
--Returns a random number between 0 and 1
--
global function TRandom()
 call(TRandom_f)
 --This conversion works, but speed is decreased!
 --This needs to be written in assembler,
 --this is why TRandom is slower than rand
 return (float64_to_atom(peek({FLOAT32,8}))-1.0)
end function

--
-- Returns a Random number between p_min and p_max
--
global function TIRandom(integer p_min,integer p_max)
  poke4(max,p_max)
  poke4(min,p_min)

  call(TIRandom_f)

 return (peek4u(returnvalue))
end function

--
-- This is my own TRandom routine. It's faster than TRandom()
-- And I encourage you to use it.
--
global function J_TRandom()
  call(J_TIRandom_f)
  return (peek4u(returnvalue)/1073741823)
end function


